﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Xml;
using System.Xml.Serialization;
using System.IO;
using System.Diagnostics;
using System.Net;

namespace AddressGeoXML
{
    public partial class AddGeoXML : Form
    {
        public AddGeoXML()
        {
            InitializeComponent();
        }


        HttpWebRequest request;
        Stream responseReader;
        Uri address;
        byte[] byteData;
        Stream postStream = null;
        String tServer;
        int recordCount = 0;
        int tempRecordCount = 1;
        int currentRecord = 1;
        XmlTextWriter Writer;
        StreamReader fp;
        String xmlData;
        XmlDocument xmlDoc;
        System.Xml.XmlDocument rootNode = new System.Xml.XmlDocument();
        System.Xml.XmlDocument rootNodeGeo = new System.Xml.XmlDocument();
        Boolean isInitializedAdd = true;
        Boolean isInitializedGeo = false;



        //***************************************************************************
        // Function to Initialize DQWS V2 XML
        //****************************************************************************
        private void InitializeAdd()
        {
            // Initialize XML Text Writer
            Writer = new XmlTextWriter("request.xml", null);

            // Write XML Header
            Writer.WriteStartDocument();
            Writer.Formatting = Formatting.Indented;
            Writer.Indentation = 3;

            Writer.WriteStartElement("RequestArray");
            Writer.WriteElementString("TransmissionReference", "urn:mdWebServiceAddress", "Testing: DQWS XML Sample Code implementation using multiple record inputs.");
            Writer.WriteElementString("CustomerID", txtCustomerID.Text);
            Writer.WriteElementString("OptAddressParsed", "True");


        }

        private void btnInsert_Click(object sender, EventArgs e)
        {
            // Execute this only if no records have been inserted yet.
            if (recordCount == 0)
            {
                //Initialize XML Text Writer
                InitializeAdd();
            }

            // Keep inserting records until btnLookup is pressed
            Writer.WriteStartElement("Record");
            Writer.WriteElementString("RecordID", Convert.ToString(recordCount + 1));
            Writer.WriteElementString("Company", txtCompany.Text);
            Writer.WriteElementString("Urbanization", txtUrbanization.Text);
            Writer.WriteElementString("AddressLine1", txtAddress1.Text);
            Writer.WriteElementString("AddressLine2", txtAddress2.Text);
            Writer.WriteElementString("Suite", txtSuite.Text);
            Writer.WriteElementString("City", txtCity.Text);
            Writer.WriteElementString("State", txtState.Text);
            Writer.WriteElementString("Zip", txtZip.Text);
            Writer.WriteElementString("Plus4", txtPlus4.Text);
            Writer.WriteElementString("Country", txtCountry.Text);
            Writer.WriteElementString("LastName", txtLastName.Text);
            Writer.WriteEndElement();

            // Increment Record Count
            recordCount = recordCount + 1;
            txtRecordCount.Text = recordCount.ToString();
        }

        private void btnLookUp_Click(object sender, EventArgs e)
        {
            // Clear the XML Each Time
            webBrowser1.Navigate("about:blank");
            webBrowser2.Navigate("about:blank");

            // If recordCount = 0 that means there are no records, then simply do the current user inputs
            if (recordCount == 0)
            {
                // Initialize XML Text Writer
                InitializeAdd();

                // Keep inserting records until btnLookup is pressed
                Writer.WriteStartElement("Record");
                Writer.WriteElementString("RecordID", Convert.ToString(recordCount + 1));
                Writer.WriteElementString("Company", txtCompany.Text);
                Writer.WriteElementString("Urbanization", txtUrbanization.Text);
                Writer.WriteElementString("AddressLine1", txtAddress1.Text);
                Writer.WriteElementString("AddressLine2", txtAddress2.Text);
                Writer.WriteElementString("Suite", txtSuite.Text);
                Writer.WriteElementString("City", txtCity.Text);
                Writer.WriteElementString("State", txtState.Text);
                Writer.WriteElementString("Zip", txtZip.Text);
                Writer.WriteElementString("Plus4", txtPlus4.Text);
                Writer.WriteElementString("Country", txtCountry.Text);
                Writer.WriteElementString("LastName", txtLastName.Text);

                Writer.WriteEndElement();

                // Increment Record to 1
                recordCount = 1;
            }

            // When the lookup button is pressed, then we can end XmlWriter
            Writer.WriteEndElement();
            Writer.Close();

            // Set the Server URL to the form input
            tServer = "https://addresscheck.melissadata.net/v2/XML/Service.svc/doAddressCheck";
            address = new Uri(tServer);

            // Create the web request 
            try
            {
                request = (HttpWebRequest)WebRequest.Create(address);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }


            request.Method = "POST";
            request.ContentType = "application/xml; charset=utf-8";

            // Read in XML file and then convert it to string
            fp = File.OpenText("request.xml");
            xmlData = fp.ReadToEnd();
            fp.Close();

            byteData = UTF8Encoding.UTF8.GetBytes(xmlData.ToString());

            // Set the content length in the request headers   
            request.ContentLength = byteData.Length;

            // Write data to a stream
            try
            {
                postStream = request.GetRequestStream();
                postStream.Write(byteData, 0, byteData.Length);
            }
            finally
            {
                if (postStream != null)
                {
                    postStream.Close();
                }
            }

            // Get response   
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();

            // Get the response stream into a reader   
            responseReader = response.GetResponseStream();

            XmlTextWriter outXML;
            outXML = new XmlTextWriter("response.xml", null);

            xmlDoc = new XmlDocument();
            xmlDoc.Load(responseReader);
            xmlDoc.Save(outXML);
            Writer.Close();
            outXML.Close();

            rootNode.Load("Response.xml");

            // Clear previous results, then re-populate fields.
            ClearResults();
            ClearResultsGeo();


            // Fill the Individual Fields. Each time verify is hit, must return record counter to 1
            // Must be initialized correctly to continue
            currentRecord = 1;


            // Perform Outputting first record onto the output fields
            // Must be initialized Correctly to continue
            isInitializedAdd = true;
            isInitializedGeo = false;
            PopulateNodes(rootNode);

            if (isInitializedAdd) 
            {
                // Enable Geocode, Next and Previous Buttons
                btnNext.Enabled = true;
                btnPrevious.Enabled = true;
                btnGeo.Enabled = true;



                webBrowser1.Navigate(System.Environment.CurrentDirectory + "\\response.xml");

                responseReader.Close();

                // Set the Number of Total Records
                tempRecordCount = recordCount;
                txtRecNo.Text = currentRecord.ToString();
                txtTotal.Text = recordCount.ToString();

                // Output current record number
                txtRecordCount.Text = recordCount.ToString();
            }
            else
            {
                btnNext.Enabled = false;
                btnPrevious.Enabled = false;

                tempRecordCount = 1;
                currentRecord = 1;

                txtTotal.Text = Convert.ToString(0);
                txtRecNo.Text = Convert.ToString(0);
            }

            // Reset Record Counts
            recordCount = 0;
            txtRecordCount.Text = recordCount.ToString();

            // Switch to the appropriate tab
            tabControl1.SelectedTab = tabAddressCheck;
            tabControl2.SelectedTab = tabAddOut;
        }



        //***************************************************************************
        // Function to Clear Output Fields
        //****************************************************************************
        private void ClearResults()
        {
            cAddress1.Clear();
            cAddress2.Clear();
            cCompany.Clear();
            cCity.Clear();
            cState.Clear();
            cZip.Clear();
            cPlus4.Clear();
            cCountry.Clear();
            cRange.Clear();
            cStreetName.Clear();
            cSuffix.Clear();
            cPreDirection.Clear();
            cPostDirection.Clear();
            cSuiteName.Clear();
            cSuiteNumber.Clear();
            cPMBName.Clear();
            cPMBNumber.Clear();
            cUName.Clear();
            cDPC.Clear();
            cDPCD.Clear();
            cCarrierRoute.Clear();
            cResultCode.Clear();
            cResultDescription.Clear();
            cInitErrCde.Clear();
            cInitErrDesc.Clear();
            cAddKey.Clear();
        }


        // Populate the Node List through a Recursive Call from the root
        public void PopulateNodes(XmlDocument xml)
        {
            foreach (XmlNode xmlnode in xml)
            {
                SearchRecord(xmlnode);
            }
        }

        // Search for Current Record In the Record List
        public void SearchRecord(XmlNode xml)
        {
            //For Each parentNode As XmlNode In xml.ChildNodes
            foreach (XmlNode xmlNode in xml)
            {

                if ((xmlNode.Name == "Results") && (xmlNode.HasChildNodes))
                {
                    // ***********************************************************************
                    // Block to Handle Different Error Codes in Initialization Results
                    String errorCodeString = xmlNode.FirstChild.Value;
                    String[] errorCodeList = null;

                    errorCodeList = errorCodeString.Split(',');

                    foreach (String errorCode in errorCodeList)
                    {

                        cInitErrCde.Text += errorCode + "; ";

                        if (errorCode == "SE01")
                        {
                            cInitErrDesc.Text += errorCode + ": Web Service Internal Error;  ";
                        }
                        else if (errorCode == "GE01")
                        {
                            cInitErrDesc.Text += errorCode + ": Empty XML Request Structure;  ";
                        }
                        else if (errorCode == "GE02")
                        {
                            cInitErrDesc.Text += errorCode + ": Empty XML Request Record Structure;  ";
                        }
                        else if (errorCode == "GE03")
                        {
                            cInitErrDesc.Text += errorCode + ": Counted records send more than number of records allowed per request;  ";
                        }
                        else if (errorCode == "GE04")
                        {
                            cInitErrDesc.Text += errorCode + ": CustomerID empty;  ";
                            MessageBox.Show("Please enter your Customer ID", "Invalid Customer ID");
                        }
                        else if (errorCode == "GE05")
                        {
                            cInitErrDesc.Text += errorCode + ": CustomerID not valid;  ";
                            MessageBox.Show("The Customer ID your entered is invalid. To retrieve a valid " +
                                 "Customer ID, please contact a Melissa Data Sales Representative " +
                                 "at 800-MELISSA ext. 3 (800-800-6245 ext. 3). ", "Invalid Customer ID");
                        }
                        else if (errorCode == "GE06")
                        {
                            cInitErrDesc.Text += errorCode + ": CustomerID disabled;  ";
                            MessageBox.Show("The Customer ID your entered has been disabled. " +
                                 "Please contact a Melissa Data Sales Representative " +
                                 "at 800-MELISSA ext. 3 (800-800-6245 ext. 3). ", "Disabled Customer ID");
                        }
                        else if (errorCode == "GE07")
                        {
                            cInitErrDesc.Text += errorCode + ": XML Request invalid;  ";
                        }
                    }

                    // Did not Initialize. Disable Buttons
                    isInitializedAdd = false;
                    return;
                }
                //***********************************************************************

                else if ((xmlNode.Name == "Results") && !(xmlNode.HasChildNodes))
                {
                    // If No Initialize Error
                    cInitErrDesc.Text = "No Error";
                }

                else if (xmlNode.Name == "Record")
                {
                    // If The RecordID is equal to the current node being selected, then output all it's sub nodes.
                    if ((xmlNode.FirstChild.Name == "RecordID") && (xmlNode.FirstChild.FirstChild.Value == currentRecord.ToString()))
                    {
                        PopulateSubNodes(xmlNode);
                    }
                }
            }
        }


        // Popoulate all Sub Nodes below the chosen Record Node
        public void PopulateSubNodes(XmlNode parentxmlnode)
        {
            foreach (XmlNode node in parentxmlnode)
            {
                Add(node);
                PopulateSubNodes(node);
            }
        }


        //***************************************************************************
        // Functions to Populate The Output Fields
        //****************************************************************************
        // Populate Output Fields that have existing nodes
        public void Add(XmlNode theNode)
        {

            if (theNode.ParentNode.Name == "Results")
            {

                //**********************************************************************************
                // Block to Handle Different Error Codes and Status Codes in each Record
                //**********************************************************************************
                String errorCodeString = theNode.Value;
                String[] errorCodeList = null;

                errorCodeList = errorCodeString.Split(',');

                foreach (String errorCode in errorCodeList)
                {

                    cResultCode.Text += errorCode + "; ";

                    if (errorCode == "AE01")
                        cResultDescription.Text += errorCode + ": Zip Code Error;  ";
                    else if (errorCode == "AE02")
                        cResultDescription.Text += errorCode + ": Unknown Street;  ";
                    else if (errorCode == "AE03")
                        cResultDescription.Text += errorCode + ": Component Error;  ";
                    else if (errorCode == "AE04")
                        cResultDescription.Text += errorCode + ": Non-Deliverable Address;  ";
                    else if (errorCode == "AE05")
                        cResultDescription.Text += errorCode + ": Address Matched to Multiple Records;  ";
                    else if (errorCode == "AE06")
                        cResultDescription.Text += errorCode + ": Address Matched to Early Warning System;  ";
                    else if (errorCode == "AE07")
                        cResultDescription.Text += errorCode + ": Empty Address Input;  ";
                    else if (errorCode == "AE08")
                        cResultDescription.Text += errorCode + ": Suite Range Error;  ";
                    else if (errorCode == "AE09")
                        cResultDescription.Text += errorCode + ": Suite Range Missing;  ";
                    else if (errorCode == "AE10")
                        cResultDescription.Text += errorCode + ": Primary Range Error;  ";
                    else if (errorCode == "AE11")
                        cResultDescription.Text += errorCode + ": Primary Range Missing;  ";
                    else if (errorCode == "AE12")
                        cResultDescription.Text += errorCode + ": Box Number Error from PO Box or RR;  ";
                    else if (errorCode == "AE13")
                        cResultDescription.Text += errorCode + ": PO Box Number Missing from PO Box or RR;  ";
                    else if (errorCode == "AE14")
                        cResultDescription.Text += errorCode + ": Input Address Matched to CMRA but secondary number not present;  ";

                    else if (errorCode == "AS01")
                        cResultDescription.Text += errorCode + ": Address matched to postal database;  ";
                    else if (errorCode == "AS02")
                        cResultDescription.Text += errorCode + ": Address matched to non-postal database;  ";
                    else if (errorCode == "AS09")
                        cResultDescription.Text += errorCode + ": Foreign Postal Code Detected;  ";
                    else if (errorCode == "AS10")
                        cResultDescription.Text += errorCode + ": Address matched to CMRA;  ";
                    else if (errorCode == "AS11")
                        cResultDescription.Text += errorCode + ": Address Vacant;  ";
                    else if (errorCode == "AS12")
                        cResultDescription.Text += errorCode + ": Address deliverable by non-USPS;  ";
                    else if (errorCode == "AS13")
                        cResultDescription.Text += errorCode + ": Address Updated By LACS;  ";
                    else if (errorCode == "AS14")
                        cResultDescription.Text += errorCode + ": Address Updated By Suite Link;  ";
                    else if (errorCode == "AS15")
                        cResultDescription.Text += errorCode + ": Address Updated By Suite Finder;  ";

                }

                if (errorCodeString == "AS01,AS12")
                {
                    cResultDescription.Clear();
                    cResultDescription.Text += "Input Address Matched to DPV;  "; ;
                }
                // End of Error Handling
                //*********************************************************************************
            }

            else if (theNode.ParentNode.Name == "Company")
                cCompany.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Address1")
                cAddress1.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Address2")
                cAddress2.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Zip")
            {
                if (theNode.ParentNode.ParentNode.Name == "Address")
                    cZip.Text = theNode.Value;
            }
            else if (theNode.ParentNode.Name == "Plus4")
                cPlus4.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "AddressRange")
                cRange.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "StreetName")
                cStreetName.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Suffix")
                cSuffix.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Pre")
                cPreDirection.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Post")
                cPostDirection.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "AddressKey")
                cAddKey.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "DeliveryPointCode")
                cDPC.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "DeliveryPointCheckDigit")
                cDPCD.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "CarrierRoute")
                cCarrierRoute.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Abbreviation")
            {
                // Several Nodes use "Code". We must handle and distinguish which Code is refering to
                if (theNode.ParentNode.ParentNode.Name == "State")
                    cState.Text = theNode.Value;
            }
            else if (theNode.ParentNode.Name == "Code")
            {
                // Several Nodes use "Code". We must handle and distinguish which Code is refering to
                if (theNode.ParentNode.ParentNode.Name == "Urbanization")
                    cUName.Text = theNode.Value;
            }
            else if (theNode.ParentNode.Name == "Range")
            {
                // Several Nodes use "Range". We must handle and distinguish which Range is refering to
                if (theNode.ParentNode.ParentNode.Name == "Suite")
                    cSuiteNumber.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "PrivateMailbox")
                    cPMBNumber.Text = theNode.Value;
            }
            else if (theNode.ParentNode.Name == "Name")
            {
                // Several Nodes use "Name". We must handle and distinguish which name is refering to
                if (theNode.ParentNode.ParentNode.Name == "Suite")
                    cSuiteName.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "City")
                    cCity.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "Urbanization")
                    cUName.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "Country")
                    cCountry.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "PrivateMailbox")
                    cPMBName.Text = theNode.Value;
            }
        }


        private void btnClear_Click(object sender, EventArgs e)
        {
            // Clear AddressCheck Fields
            txtAddress1.Text = "";
            txtAddress2.Text = "";
            txtCity.Text = "";
            txtPlus4.Text = "";
            txtZip.Text = "";
            txtState.Text = "";
            txtCompany.Text = "";
            txtSuite.Text = "";
            txtUrbanization.Text = "";
            txtCountry.Text = "";
        }


        //######################################################################################
        //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        // GEOCODE FUNCTIONS
        //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        //######################################################################################



        //***************************************************************************
        // Function to Clear Output Fields
        //****************************************************************************
        private void ClearResultsGeo()
        {
            cCountyName.Clear();
            cCountyFips.Clear();
            cLat.Clear();
            cLong.Clear();
            cPlaceCode.Clear();
            cPlaceName.Clear();
            cPlaceName.Clear();
            cTZN.Clear();
            cTZC.Clear();
            cCBSACode.Clear();
            cCBSATitle.Clear();
            cCBSALevel.Clear();
            cCBSADCode.Clear();
            cCBSADTitle.Clear();
            cCBSADLevel.Clear();
            cCensusBlock.Clear();
            cCensusTract.Clear();
            cResultGeo.Clear();
            cResultDescGeo.Clear();
            cInitErrGeo.Clear();
            cInitErrDescGeo.Clear();
        }


        private void InitializeGeo()
        {
            // Initialize XML Text Writer
            Writer = new XmlTextWriter("request.xml", null);

            // Write XML Header
            Writer.WriteStartDocument();
            Writer.Formatting = Formatting.Indented;
            Writer.Indentation = 3;

            Writer.WriteStartElement("RequestArray");
            Writer.WriteElementString("TransmissionReference", "urn:mdWebServiceAddress", "Testing: DQWS XML Sample Code implementation using multiple record inputs.");
            Writer.WriteElementString("CustomerID", txtCustomerID.Text);
        }


        private void btnGeo_Click(object sender, EventArgs e)
        {
            InitializeGeo();


            int N;

            for(N = 0; N < Convert.ToInt32(txtTotal.Text); N++)
            {

            String Temp;

                if (rootNode.GetElementsByTagName("AddressKey").Item(N).HasChildNodes == true)
                {
                    Temp = rootNode.GetElementsByTagName("AddressKey").Item(N).FirstChild.Value;
                }
                else
                {
                    Temp = "";
                }
            
                // Keep inserting records until btnLookup is pressed
                Writer.WriteStartElement("Record");
                Writer.WriteElementString("RecordID", Convert.ToString(N + 1));
                Writer.WriteElementString("AddressKey", Temp);


                Writer.WriteEndElement();
            }

            recordCount = Convert.ToInt32(txtTotal.Text);

            // When the lookup button is pressed, then we can end XmlWriter
            Writer.WriteEndElement();
            Writer.Close();

            // Set the Server URL to the form input
            tServer = "https://geocoder.melissadata.net/v2/XML/Service.svc/doGeoCode";
            address = new Uri(tServer);

            // Create the web request 
            try
            {
                request = (HttpWebRequest)WebRequest.Create(address);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }


            request.Method = "POST";
            request.ContentType = "application/xml; charset=utf-8";

            // Read in XML file and then convert it to string
            fp = File.OpenText("request.xml");
            xmlData = fp.ReadToEnd();
            fp.Close();

            byteData = UTF8Encoding.UTF8.GetBytes(xmlData.ToString());

            // Set the content length in the request headers   
            request.ContentLength = byteData.Length;

            // Write data to a stream
            try
            {
                postStream = request.GetRequestStream();
                postStream.Write(byteData, 0, byteData.Length);
            }
            finally
            {
                if (postStream != null)
                {
                    postStream.Close();
                }
            }

            // Get response   
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();

            // Get the response stream into a reader   
            responseReader = response.GetResponseStream();

            XmlTextWriter outXML;
            outXML = new XmlTextWriter("response.xml", null);

            xmlDoc = new XmlDocument();
            xmlDoc.Load(responseReader);
            xmlDoc.Save(outXML);
            Writer.Close();
            outXML.Close();


            // Clear previous results, then re-populate fields.
            ClearResultsGeo();

            rootNodeGeo.Load("Response.xml");

            // Perform Outputting first record onto the output fields
            // Must be initialized Correctly to continue
            isInitializedGeo = true;
            PopulateNodesGeo(rootNodeGeo);

            if (isInitializedGeo)
            {

                btnNext.Enabled = true;
                btnPrevious.Enabled = true;

                // Fill the Individual Fields. Each time verify is hit, must return record counter to 1
                currentRecord = 1;

                // View response xml
                webBrowser2.Navigate(System.Environment.CurrentDirectory + "\\response.xml");

                responseReader.Close();

                // Set records back to 0. Everytime the records are processed, XML requests must be redone.
                tempRecordCount = recordCount;
                txtRecNo.Text = currentRecord.ToString();
                //txtTotal.Text = recordCount
                recordCount = 0;
                txtRecordCount.Text = recordCount.ToString();
            }
            else
            {
                // Set the Number of Total Records
                //TotalRecords = 1
                currentRecord = 1;
                txtTotal.Text = Convert.ToString(0);
                txtRecNo.Text = Convert.ToString(0);

            }

            // Reset Record Counts
            recordCount = 0;
            txtRecordCount.Text = recordCount.ToString();

            // Switch to the appropriate tab
            tabControl1.SelectedTab = tabGeoCode;
            tabControl3.SelectedTab = tabGeoOut;

            // Disable Geocode Button after doing GeoCode
            btnGeo.Enabled = false;

        }


        // ***************************************************************************
        // Functions to Populate The Output Fields
        //****************************************************************************
        // Populate Output Fields that have existing nodes
        public void AddGeo(XmlNode theNode)
        {
            if (theNode.ParentNode.Name == "Results")
            {

                //**********************************************************************************
                // Block to Handle Different Error Codes and Status Codes in each Record
                //**********************************************************************************
                String errorCodeString = theNode.Value;
                String[] errorCodeList = null; 

                errorCodeList = errorCodeString.Split(',');
               
                foreach (String errorCode in errorCodeList)
                {
                    cResultGeo.Text += errorCode + "; ";

                    if (errorCode == "GS01")
                        cResultDescGeo.Text += errorCode + ": Geocoded to Zip+4 or the full 6-digit Postal Code Centroid;  ";
                    else if (errorCode == "GS02")
                        cResultDescGeo.Text += errorCode + ": Geocoded to Zip+2 Centroid;  ";
                    else if (errorCode == "GS03")
                        cResultDescGeo.Text += errorCode + ": Geocoded to 5 digit (U.S.) or 3 digit (Canada) zip code Centroid;  ";
                    else if (errorCode == "GS05")
                        cResultDescGeo.Text += errorCode + ": Geocoded to Rooftop level;  ";
                    else if (errorCode == "GS06")
                        cResultDescGeo.Text += errorCode + ": Geocoded to Interpolated Rooftop level;  ";
                    else if (errorCode == "GE01") 
                        cResultDescGeo.Text += errorCode + ": Zip Code Error;  ";
                    else if (errorCode == "GE02")
                    {
                        cResultDescGeo.Text += errorCode + ": Error in Data Field (Empty AddressKey field," +
                              "Address length < 11, or Non-digit encountered;  ";
                    }

                }

                // End of Error Handling
                //*********************************************************************************
            }

            else if (theNode.ParentNode.Name == "Fips") 
                cCountyFips.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Latitude") 
                cLat.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Longitude") 
                cLong.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Title") 
                cCBSATitle.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Level") 
                cCBSALevel.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "CBSADivisionCode") 
                cCBSADCode.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "CBSADivisionTitle") 
                cCBSADTitle.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "CBSADivisionLevel") 
                cCBSADLevel.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Block") 
                cCensusBlock.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Tract") 
                cCensusTract.Text = theNode.Value;
            else if (theNode.ParentNode.Name == "Code") 
            {
                // Several Nodes use "Range". We must handle and distinguish which Range is refering to
                if (theNode.ParentNode.ParentNode.Name == "Place") 
                    cPlaceCode.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "TimeZone") 
                    cTZC.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "CBSA") 
                    cCBSACode.Text = theNode.Value;
            }
            else if (theNode.ParentNode.Name == "Name") 
            {
                // Several Nodes use "Name". We must handle and distinguish which name is refering to
                if (theNode.ParentNode.ParentNode.Name == "County") 
                    cCountyName.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "Place") 
                    cPlaceName.Text = theNode.Value;
                else if (theNode.ParentNode.ParentNode.Name == "TimeZone") 
                    cTZN.Text = theNode.Value;
                
            }
        }

        // Populate the Node List through a Recursive Call from the root
        public void PopulateNodesGeo(XmlDocument xml)
        {
            foreach (XmlNode xmlnode in xml)
            {
                SearchRecordGeo(xmlnode);
            }
        }


        // Search for Current Record In the Record List
        public void SearchRecordGeo(XmlNode xml)
        {
            //For Each parentNode As XmlNode In xml.ChildNodes
            foreach (XmlNode xmlNode in xml)
            {

                if ((xmlNode.Name == "Results") && (xmlNode.HasChildNodes))
                {
                    //***********************************************************************
                    // Block to Handle Different Error Codes in Initialization Results
                    String errorCodeString = xmlNode.FirstChild.Value;
                    String[] errorCodeList = null;

                    errorCodeList = errorCodeString.Split(',');

                    foreach (String errorCode in errorCodeList)
                    {

                        cInitErrGeo.Text += errorCode + "; ";

                        if (errorCode == "SE01") 
                            cInitErrDescGeo.Text += errorCode + ": Web Service Internal Error;  ";
                        else if (errorCode == "GE01") 
                            cInitErrDescGeo.Text += errorCode + ": Empty XML Request Structure;  ";
                        else if  (errorCode == "GE02") 
                            cInitErrDescGeo.Text += errorCode + ": Empty XML Request Record Structure;  ";
                        else if (errorCode == "GE03") 
                            cInitErrDescGeo.Text += errorCode + ": Counted records send more than number of records allowed per request;  ";
                        else if  (errorCode == "GE04") 
                        {
                            cInitErrDescGeo.Text += errorCode + ": CustomerID empty;  ";
                            MessageBox.Show("Please enter your Customer ID", "Invalid Customer ID");
                        }
                        else if (errorCode == "GE05") 
                        {
                            cInitErrDescGeo.Text += errorCode + ": CustomerID not valid;  ";
                            MessageBox.Show("The Customer ID your entered is invalid. To retrieve a valid " +
                                  "Customer ID, please contact a Melissa Data Sales Representative " +
                                  "at 800-MELISSA ext. 3 (800-800-6245 ext. 3). ", "Invalid Customer ID");
                        }
                        else if (errorCode == "GE06") 
                        {
                            cInitErrDescGeo.Text += errorCode + ": CustomerID disabled;  ";
                            MessageBox.Show("The Customer ID your entered has been disabled. " +
                                  "Please contact a Melissa Data Sales Representative " +
                                  "at 800-MELISSA ext. 3 (800-800-6245 ext. 3). ", "Disabled Customer ID");
                        }
                        else if (errorCode == "GE07") 
                            cInitErrDescGeo.Text += errorCode + ": XML Request invalid;  ";

                    }
                    //***********************************************************************
                }
                else if  ((xmlNode.Name == "Results") && !(xmlNode.HasChildNodes))
                {
                    // If No Initialize Error
                    cInitErrDescGeo.Text = "No Error";
                }
                else if  (xmlNode.Name == "Record")
                {
                    // If The RecordID is equal to the current node being selected, then output all it's sub nodes.
                    if ((xmlNode.FirstChild.Name == "RecordID") && (xmlNode.FirstChild.FirstChild.Value == currentRecord.ToString()))
                        PopulateSubNodesGeo(xmlNode);
                }

            }
        }

        // Popoulate all Sub Nodes below the chosen Record Node
        public void PopulateSubNodesGeo(XmlNode parentxmlnode)
        {
            foreach (XmlNode node in parentxmlnode)
            {
                AddGeo(node);
                PopulateSubNodesGeo(node);
            }
        }

        private void btnPrevious_Click(object sender, EventArgs e)
        {
            // Clear Fields Each Time
            ClearResults();
            ClearResultsGeo();

            // Decrement the Record Counter
            if (currentRecord != 1)
                currentRecord = currentRecord - 1;
            else
                currentRecord = tempRecordCount;
            
            // Fill the Individual Fields
            PopulateNodes(rootNode);

            if (isInitializedGeo)
                PopulateNodesGeo(rootNodeGeo);
       

            txtRecNo.Text = currentRecord.ToString();
        }

        private void btnNext_Click(object sender, EventArgs e)
        {
            // Clear Fields Each Time
            ClearResults();
            ClearResultsGeo();

            // Increment the Record Counter
            if (currentRecord != tempRecordCount)
                currentRecord = currentRecord + 1;
            else
                currentRecord = 1;

            // Fill the Individual Fields
            PopulateNodes(rootNode);

            if (isInitializedGeo)
                PopulateNodesGeo(rootNodeGeo);

            txtRecNo.Text = currentRecord.ToString();
        }     

    }
}
